-- Puffin: A Tojaso FowlPlay Project
-- Adapter component for data providers related to units

-- Providers are used to access values of various classes that are derived from information in data records.
-- Providers can be assigned during configuration and later used to get current values during updates.
-- Providers are created and accessed through the framework database.
-- To create or access a provider, you need to know the class of provider (e.g., "color", "text") and
-- the name of a specific provider (e.g., "unit_health_percent", "class_color").
--		DB.SetAttribute("providers", "class", "name", v)
--		v = DB.GetAttribute("providers", "class", "name")
-- The argument and/or returned value "v" is either a handle on a function or a value of the type associated with the provider.
-- If the provider is a function then it is called with appropriate arguments (this can be overridden by including
-- "no_function = true" in the provider reference table).
-- Provider functions defined in this adapter expect the data record for an active unit as the only argument).
-- If "v" is not a function then it will be of the type associated with the provider and can be used directly
-- (this is typically used to access constants such as colors in a color palette).
-- Special case: class = "condition" name = "exists" just tests for existence of the unit
-- A provider function can return a table containing more than one attribute (the code that processes the result of a
-- provider function call determines this case by the existence of a key in the returned table that matches the attribute
-- name associated with the provider reference table in a configuration file. This can be overridden by including
-- "no_table = true" in the provider reference table.

local Adapter_UnitProviders = {}
local EF = Puffin.ErrorFormat
local DB

-- Convert color codes from hex number to array with r, g, b, a fields (alpha set to 1.0)
local function HexColor(x)
	local n = tonumber(x, 16)
	local r, g, b = math.floor(n / (256 * 256)), math.floor(n / 256) % 256, n % 256
	return { r = r / 255, g = g / 255, b = b / 255, a = 1 }
end

-- Colors for classes
local cc = {
	DEATHKNIGHT = HexColor("C41F3B"), DRUID = HexColor("FF7D0A"), HUNTER = HexColor("ABD473"), MAGE = HexColor("69CCF0"),
	PALADIN = HexColor("F58CBA"), PRIEST = HexColor("FFFFFF"), ROGUE = HexColor("FFF569"), SHAMAN = HexColor("0070DE"),
	WARLOCK = HexColor("9482C9"), WARRIOR = HexColor("C79C6E"), MONK = HexColor("558A84"),
}
local ccb = {} -- Background colors that coordinate with the class colors

local classificationText = { normal = "", elite = "Elite ", rare = "Rare ", rareelite = "Rare-Elite ", worldboss = "Boss " }

-- Return a shortened number string
local function NumberToString(v)
	local f, n = "%d", math.abs(v)
	if n >= 10000000 then f = "%.1fm"; v = v / 1000000
	elseif n >= 1000000 then f = "%.2fm"; v = v / 1000000
	elseif n >= 100000 then f = "%.0fk"; v = v / 1000
	elseif n >= 10000 then f = "%.1fk"; v = v / 1000
	else v = math.floor(v + 0.5) end
	return string.format(f, v)
end

local function NumberAsPercent(v) return string.format("%.0f", math.abs(v) * 100) end
local function NumberWithPercent(v) return string.format("%.0f%%", math.abs(v) * 100) end
local function NumberSlash(v1, v2) if v2 > 0 then return string.format("%s/%s", NumberToString(v1), NumberToString(v2)) else return "" end end 
local function NumberSlashPercent(v1, v2) return string.format("%s/%s %s", NumberToString(v1), NumberToString(v2), NumberWithPercent(v1/v2)) end 
local function NumberPercent(v1, v2) return string.format("%s %s", NumberToString(v1), NumberWithPercent(v1/v2)) end 


local function HealthMax(t) if t.healthMax then return NumberSlash(t.health or 0, t.healthMax) else return "" end end
local function HealthMaxPercent(t) if t.healthMax then return NumberSlashPercent(t.health or 0, t.healthMax) else return "" end end
local function HealthPercent(t) if t.healthMax then return NumberPercent(t.health or 0, t.healthMax) else return "" end end

local function IncomingHeals(t)
	if t.healthIncoming and t.healthMax and t.healthIncoming > (t.healthMax * 0.01) then return "+" .. NumberToString(t.healthIncoming) end
end

local function PowerMax(t) if t.powerMax then return NumberSlash(t.power or 0, t.powerMax) else return "" end end
local function PowerPercent(t) if t.powerMax then return NumberPercent(t.power or 0, t.powerMax) else return "" end end

local function PowerStatus(t)
	if t.resurrection then return "Rez!" elseif t.isGhost then return "Ghost" elseif t.isDead then return "Dead" elseif not t.isConnected then return "Offline"
		elseif t.isAFK then return "AFK" elseif t.isDND then return "DND" elseif t.isFeign then return "Feign" elseif t.isPossessed then return "Possessed"
		elseif t.isCharmed then return "Charmed" end
	if t.powerMax and t.powerType then
		if t.powerType == 0 or t.powerType == "MANA" then return NumberPercent(t.power or 0, t.powerMax) else
			if t.power > 0 and t.power ~= t.powerMax then return NumberToString(t.power or 0) else return "" end end
	else return "" end
end

local function PowerStatusPercent(t)
	if t.resurrection then return "Rez!" elseif t.isGhost then return "Ghost" elseif t.isDead then return "Dead" elseif not t.isConnected then return "Offline"
		elseif t.isAFK then return "AFK" elseif t.isDND then return "DND" elseif t.isFeign then return "Feign" elseif t.isPossessed then return "Possessed"
		elseif t.isCharmed then return "Charmed" end
	if t.powerMax and t.powerType then
		if t.powerType == 0 or t.powerType == "MANA" then
			if t.power and t.power > 0 and t.powerMax and t.powerMax > 0 and t.power < t.powerMax then return NumberAsPercent(t.power / t.powerMax) end
		else
			if t.power > 0 and t.power ~= t.powerMax then return NumberToString(t.power or 0) end
		end
	end
	return ""
end

local function Status(t)
	if t.resurrection then return "Rez!" elseif t.isGhost then return "Ghost" elseif t.isDead then return "Dead" elseif not t.isConnected then return "Offline"
		elseif t.isAFK then return "AFK" elseif t.isDND then return "DND" elseif t.isFeign then return "Feign" elseif t.isPossessed then return "Possessed"
		elseif t.isCharmed then return "Charmed" end
	return ""
end

local function UnitName(t) return t.name or "Unknown" end
local function TalentTree(t) return t.talent or "" end
local function PVPTimer(t) local c = t.pvpTimer if c then c = math.floor(c); return string.format("%0d:%02d", c / 60, c %  60) else return "" end end

local function LevelClass(t)
	local lt = ""
	if t.level then
		local c = DB.GetAttribute("defaults", "colors", "level")
		if (t.level < 0) then
			lt = "|cffff0000??|r "
		elseif c then
			lt = string.format("|cff%02x%02x%02x%s|r ", c.r * 255, c.g * 255, c.b * 255, t.level)
		else
			lt = tostring(t.level) .. " "
		end
	end
	local lr = t.race or "Unknown"
	local c = DB.GetAttribute("defaults", "colors", "race")
	if c then lr = string.format("|cff%02x%02x%02x%s|r", c.r * 255, c.g * 255, c.b * 255, lr) end
	if t.isPlayer and t.classToken then
		local c = cc[t.classToken]
		return string.format("%s|cff%02x%02x%02x%s|r %s", lt, c.r * 255, c.g * 255, c.b * 255, t.class or "Unknown", lr)
	else
		local e = classificationText[t.classification or "normal"]
		local s = string.format("%s%s%s", e or "", lt or "", t.creatureType or t.creatureFamily or "Unknown")
		if t.classBase then s = s .. " " .. t.classBase end -- add class for the NPC if known
		return s
	end
end

-- Table of combo point text with embedded image escape sequences
local combo_point_text = "|TInterface\\AddOns\\Puffin\\Images\\Combo:12:12|t"
local combo_text = {}

local function ComboPoints(t)
	if not t.comboPoints then return "" else return combo_text[t.comboPoints] end
end

-- Check if a particular aura is active, return true and both timeLeft and expiration time if it is active
-- Optionally check if the aura was cast by the player
local function AuraLookup(t, name, isBuff, castByPlayer)
	if t.auras then
		local at = t.auras.buffs; if not isBuff then at = t.auras.debuffs end
		if at and at.cache and at.cache[name] then -- make sure the aura is active in the right table
			for i = 1, at.count do
				local b = at[i]
				if b.name == name and (not castByPlayer or b.caster == "player") then
					local tl = math.floor(b.expire - GetTime()); if tl < 0 then tl = 0 end
					return b, tl, b.expire
				end
			end
		end
	end
	return nil, 0, nil
end

local function ClassTime1(t)
	local class, expire, _ = Puffin.myClass, nil
	if class == "DRUID" then _, _, expire = AuraLookup(t, "Lifebloom", true, true) end
	if class == "PRIEST" then _, _, expire = AuraLookup(t, "Power Word: Shield", true, true) end
	return expire
end

local function ClassTime2(t)
	local class, expire, _ = Puffin.myClass, nil
	if class == "DRUID" then _, _, expire = AuraLookup(t, "Rejuvenation", true, true) end
	if class == "PRIEST" then _, _, expire = AuraLookup(t, "Renew", true, true) end
	return expire
end

local function ClassTime3(t)
	local class, expire, _ = Puffin.myClass, nil
	if class == "DRUID" then _, _, expire = AuraLookup(t, "Regrowth", true, true) end
	if class == "PRIEST" then _, _, expire = AuraLookup(t, "Prayer of Mending", true, true) end
	return expire
end

local function ClassTime4(t)
	local class, expire, _ = Puffin.myClass, nil
	if class == "DRUID" then _, _, expire = AuraLookup(t, "Wild Growth", true, true) end
	if class == "PRIEST" then _, _, expire = AuraLookup(t, "Glyph of Prayer of Healing", true, true) end
	return expire
end

-- Table of text-related providers
local unit_text = {
	health_max = HealthMax, health_max_percent = HealthMaxPercent, health_percent = HealthPercent, incoming_heals = IncomingHeals,
	power_max = PowerMax, power_percent = PowerPercent, power_status = PowerStatus, power_status_percent = PowerStatusPercent, status = Status,
	name = UnitName, level_class = LevelClass, pvp_timer = PVPTimer, combo_points = ComboPoints,
}

-- Return a copy of a color
local function CopyColor(c) return { r = c.r, g = c.g, b = c.b, a = c.a } end

-- Set saturation and brightness of RGB colors by converting into HSL, adjusting saturation, then converting back to RGB
-- Input color RGB components are values between 0 to 1.0, saturation and brightness are values between -1.0 and 1.0
local function LevelAdjust(v, a) -- apply adjustment in range -1..+1 to either saturation or brightness
	if a ~= 0 then if a >= -1 and a < 0 then return v * (a + 1) elseif a > 0 and a <= 1 then return v + ((1 - v) * a) end end
	return v
end

local function AdjustColor(r, g, b, saturation, brightness)
	if not r or not g or not b then return 0.5, 0.5, 0.5 end -- avoid errors if passed in nil values
	if not saturation then saturation = 0 end; if not brightness then brightness = 0 end -- set to default values
	if (saturation == 0) and (brightness == 0) then return r, g, b end	
	local ch, cs, cl, v, m, sv, h, sextant, fract, mid1, mid2, r2, g2, b2, vsf
	local mincolor, maxcolor = math.min(r, g, b), math.max(r, g, b)
	if mincolor == maxcolor then
		ch, cs, cl = 0, 0, r
	else
		v = maxcolor - mincolor; cl = (mincolor + maxcolor) / 2	
		if cl < 0.5 then cs = v / (maxcolor + mincolor) else cs = v / (2 - maxcolor - mincolor) end	
		r2 = (maxcolor - r) / v; g2 = (maxcolor - g) / v; b2 = (maxcolor - b) / v	
		if r == maxcolor then
			if g == mincolor then ch = b2 + 5 else ch = 1 - g2 end
		elseif g == maxcolor then
			if b == mincolor then ch = r2 + 1 else ch = 3 - b2 end
		else
			if r == mincolor then ch = g2 + 3 else ch = 5 - r2 end
		end		
		ch = ch / 6
	end
	if saturation < -1 then saturation = -1 elseif saturation > 1 then saturation = 1 end
	cs = LevelAdjust(cs, saturation) -- adjust the saturation, using original -1 .. +1 scale
	if brightness < -1 then brightness = -1 elseif brightness > 1 then brightness = 1 end
	cl = LevelAdjust(cl, brightness / 2) -- adjust the brightness, restricting the range to -0.5 to +0.5
	r, g, b = cl, cl, cl
	if cl <= 0.5 then v = cl * (1 + cs) else v = (cl + cs) - (cl * cs) end
	if v > 0 then
		m = cl + cl - v; sv = (v - m) / v; h = ch * 6
		sextant = math.floor(h); fract = h - sextant; vsf = v * sv * fract
		mid1 = m + vsf; mid2 = v - vsf
		if sextant == 0 then
			r, g, b = v, mid1, m
		elseif sextant == 1 then
			r, g, b = mid2, v, m
		elseif sextant == 2 then
			r, g, b = m, v, mid1
		elseif sextant == 3 then
			r, g, b = m, mid2, v
		elseif sextant == 4 then
			r, g, b = mid1, m, v
		elseif sextant == 5 then
			r, g, b = v, m, mid2
		else
			r, g, b = v, mid1, m
		end
	end
	return r, g, b
end

-- Built-in color palette, including primary colors plus variants that work together (e.g., red1 coordinates with green1 and blue1)
local cp = {
	red = { r = 1, g = 0, b = 0, a = 1 }, red1 = HexColor("ef2929"), red2 = HexColor("cc0000"), red3 = HexColor("a40000"),
	green = { r = 0, g = 1, b = 0, a = 1 }, green1 = HexColor("8ae234"), green2 = HexColor("73d216"), green3 = HexColor("4e9a06"),
	blue = { r = 0, g = 0, b = 1, a = 1 }, blue1 = HexColor("729fcf"), blue2 = HexColor("3465a4"), blue3 = HexColor("204a87"),
	yellow = { r = 1, g = 1, b = 0, a = 1 }, yellow1 = HexColor("fce94f"), yellow2 = HexColor("edd400"), yellow3 = HexColor("c4a000"),
	purple = HexColor("800080"), purple1 = HexColor("ad7fa8"), purple2 = HexColor("75507b"), purple3 = HexColor("5c3566"), purplepink = HexColor("f040ff"),
	cyan = { r = 0, g = 1, b = 1, a = 1 }, cyan1 = HexColor("97FFFF"), cyan2 = HexColor("00EEEE"), cyan3 = HexColor("00CDCD"),
	orange = HexColor("FFA500"), orange1 = HexColor("fcaf3e"), orange2 = HexColor("f57900"), orange3 = HexColor("ce5c00"), orange4 = HexColor("ff5050"),
	brown = HexColor("A52A2A"), brown1 = HexColor("e9b96e"), brown2 = HexColor("c17d11"), brown3 = HexColor("8f5902"),
	pink = HexColor("ff6eb4"), magenta = { r = 1, g = 0, b = 1, a = 1 }, cleanse1 = HexColor("9370DB"), cleanse2 = HexColor("5D478B"),
	gray = HexColor("888888"), gray1 = HexColor("e0e0e0"), gray2 = HexColor("a8a8a8"), gray3 = HexColor("4A4A4A"), darkgray = HexColor("202020"),
	black = { r = 0, g = 0, b = 0, a = 1 }, white = { r = 1, g = 1, b = 1, a = 1 },	transparent = { r = 1, g = 1, b = 1, a = 0 },
	border = { r = 0.25, g = 0.25, b = 0.25, a = 0.75 }, highlight = { r = 0.5, g = 0.5, b = 0.5, a = 0.75 },
}

-- Colors for different health situations (credit to Pitbull4 for the algorithm)
local hc = { dead = cp.gray2, dc = cp.gray1, tap = cp.gray, maxh = cp.green, h = cp.yellow, minh = cp.red }
local hcb = {} -- contains background versions of the health colors
local hct = {} -- contains one hundred gradations of health color
local hctb = {} -- contains one hundred gradations of health background

local function AdjustHealthColor(h)
	local ct = {}
	if h >= 0.5 then c1 = hc.maxh; c0 = hc.h; h = h * 2 - 1 else c1 = hc.h; c0 = hc.minh; h = h * 2 end
	ct.r = c0.r * (1 - h) + c1.r * h; ct.g = c0.g * (1 - h) + c1.g * h; ct.b = c0.b * (1 - h) + c1.b * h
	return ct -- only use this function during initialization or will bleed memory
end

local function HealthColor(t)
	if not t.isConnected then return hc.dc elseif t.isDead then return hc.dead elseif t.isTapped and not t.isTappedByPlayer then return hc.tap end
	if not t.healthMax then return hc.dead end
	if Puffin.myClass == "DRUID" and t.isPlayer and t.auras and t.auras.debuffs then
		local at = t.auras.debuffs.atypes
		if at and (at.Poison or at.Magic or at.Curse) then return cp.cleanse1 end
	end
	if Puffin.myClass == "PRIEST" and t.isPlayer and t.auras and t.auras.debuffs then
		local at = t.auras.debuffs.atypes
		if at then if at.Magic then return cp.cleanse1 elseif at.Disease then return cp.brown2 end end
	end
	return hct[math.floor(((t.health or 0) / t.healthMax) * 49.99) + 1]
end

local function HealthBackground(t)
	if not t.isConnected then return hcb.dc elseif t.isDead then return hcb.dead elseif t.isTapped and not t.isTappedByPlayer then return hcb.tap end
	if not t.healthMax then return hcb.dead end
	if Puffin.myClass == "DRUID" and t.isPlayer and t.auras and t.auras.debuffs then
		local at = t.auras.debuffs.atypes
		if at and (at.Poison or at.Magic or at.Curse) then return cp.cleanse2 end
	end
	if Puffin.myClass == "PRIEST" and t.isPlayer and t.auras and t.auras.debuffs then
		local at = t.auras.debuffs.atypes
		if at then if at.Magic then return cp.cleanse2 elseif at.Disease then return cp.brown3 end end
	end
	return hctb[math.floor(((t.health or 0) / t.healthMax) * 49.99) + 1]
end

-- Colors for different power situations
local pc = {}
local pcb = {}

local function PowerColor(t) local token = t.powerType or 0; return pc[token] end
local function PowerBackground(t) local token = t.powerType or 0; return pcb[token] end

-- Special color based on power type and status
local function PowerStatusColor(t)
	if t.isGhost or t.isDead then return cp.gray1 elseif not t.isConnected or t.isAFK or t.isDND then return cp.red elseif t.isPossessed then return cp.purple1
	elseif t.isCharmed then return cp.pink elseif t.isFeign then return cp.green2 end
	local token = t.powerType or 0; return pc[token]
end

-- Special color based on power type and status
local function StatusColor(t)
	if t.isGhost or t.isDead then return cp.gray1 elseif not t.isConnected or t.isAFK or t.isDND then return cp.red elseif t.isPossessed then return cp.purple1
	elseif t.isCharmed then return cp.pink elseif t.isFeign then return cp.green2 end
	return cp.white
end

-- Colors for class foreground and background
local function ClassColor(t) return t.isPlayer and cc[t.classToken] or cp.white end
local function ClassBackground(t) return t.isPlayer and ccb[t.classToken] or cp.black end

-- Color based on reaction, default is neutral
local reactionColor = {} -- temporary table, caution required to not use overwritten values
local function ReactionColor(t, uid) if UnitReaction("player", uid) == 5 then return cp.green end local c = reactionColor; c.r, c.g, c.b, c.a = UnitSelectionColor(uid); return c end

local function ClassReaction(t, uid) if t.isPlayer then return ClassColor(t) else return ReactionColor(t, uid) end end

-- Colors for threat situation
local threatColors = { HexColor("B0B0B0"), HexColor("FFFF77"), HexColor("FF9900"), HexColor("ff0000") }
local function ThreatColor(t, unit, color) local ts, c = t.threatSituation, color or cp.black; if ts and (ts > 0) then c = threatColors[ts + 1] end return c end
local function AggroColor(t, unit, color) if t.threatSituation == 3 or t.threatSituation == 2 then return threatColors[4] else return color or cp.black end end

local function MouseoverColor(t)
	local c = DB.GetAttribute("defaults", "colors", "mouseover")
	if t.isMouseover then return (c or cp.highlight) else return cp.transparent end
end

-- Colors for debuffs
local function DebuffColor(t) -- return colors for standard debuff types
	if t.auras and t.auras.debuffs then
		local at = t.auras.debuffs.atypes
		if at then
			if at.Poison then return cp.green end
			if at.Curse then return cp.purple end
			if at.Disease then return cp.brown end
			if at.Magic then return cp.blue end
		end
	end
	return cp.purple -- nil
end

local function ClassColor1(t)
	local class = Puffin.myClass
	if class == "DRUID" then
		local b = AuraLookup(t, "Lifebloom", true, true)
		if b then
			if b.count == 1 then return cp.orange4 elseif b.count == 2 then return cp.yellow end
		end
		return cp.green
	end
	return cp.orange
end

local function ClassColor2(t) return cp.purplepink end
local function ClassColor3(t) return cp.yellow end
local function ClassColor4(t) return cp.blue2 end
local function ClassColor5(t) local b = AuraLookup(t, "Thorns", true, false) if b then return cp.brown3 end return cp.blue3 end
local function ClassColor6(t) return cp.gray end
local function ClassColor7(t) return cp.pink end

-- Table of color-related providers
local unit_color = {
	health = HealthColor, health_background = HealthBackground, power = PowerColor, power_background = PowerBackground, power_status = PowerStatusColor,
	class = ClassColor, class_background = ClassBackground, class_reaction = ClassReaction, threat = ThreatColor, aggro = AggroColor,
	mouseover = MouseoverColor, status = StatusColor, debuff = DebuffColor, reaction = ReactionColor,
	class_color1 = ClassColor1, class_color2 = ClassColor2, class_color3 = ClassColor3, class_color4 = ClassColor4,
	class_color5 = ClassColor5, class_color6 = ClassColor6, class_color7 = ClassColor7,
}

-- Health calculation, must be between 0 and 1 and be the fraction health/healthMax
local function HealthFraction(t)
	local h, hm = (t.health or 0) + 0.5, t.healthMax or 0
	if t.isDead or hm <= 0 or h <= 0 then return 0 end
	if h > hm then h = hm end
	return h / hm
end

-- Power calculation, must be between 0 and 1 and be the fraction power/powerMax
local function PowerFraction(t)
	local p, pm = (t.power or 0) + 0.5, t.powerMax or 0
	if t.isDead or pm <= 0 or p <= 0 then return 0 end
	if p > pm then p = pm end
	return p / pm
end

local function RangeAlpha(t) if t.inRange then return 1 end return DB.GetAttribute("defaults", "number", "out_of_range_alpha") or 0.5 end
local function DeadSparkAlpha(t) if t.isDead then return 0 end return 1 end
local function Direction(t) return t.direction end

local function NotManaOffset(t, unit, base) if t.powerType == 0 or t.powerType == "MANA" or
	t.isGhost or t.isDead or not t.isConnected or t.isAFK or t.isDND or t.isFeign or t.isPossessed or t.isCharmed then return base end return base + 4 end

local function IncomingFraction(t)
	local i, h, hm = (t.healthIncoming or 0), (t.health or 0), t.healthMax or 0
	if hm <= 0 then return 0 end
	if (i + h) > hm then i = hm - h end
	return i / hm
end

-- Table of numeric providers
local unit_number = {
	health_fraction = HealthFraction, power_fraction = PowerFraction, incoming_fraction = IncomingFraction, not_mana_offset = NotManaOffset,
	range_alpha = RangeAlpha, dead_spark_alpha = DeadSparkAlpha, direction = Direction,
	class_time1 = ClassTime1, class_time2 = ClassTime2, class_time3 = ClassTime3, class_time4 = ClassTime4,
}

-- Condition tests, return true or false
local function PlayerResting(t) return t.isResting == 1 end
local function PlayerPVP(t) return (t.pvp == 1) or (t.pvpFFA == 1) end
local function RaidIconCheck(t) return t.raidTarget ~= nil end
local function AssignedRole(t) return (t.role ~= nil) and (t.role ~= "NONE") end
local function Leader(t) return (t.inParty ~= nil and t.isPartyLeader ~= nil) or (t.inRaid == true and t.isRaidLeader == true) end
local function MasterLooter(t) return t.isMasterLooter end
local function ReadyCheckActive(t) return t.ready ~= nil end
local function HasComboPoints(t) return (t.comboPoints ~= nil) and (t.comboPoints >= 1) end
local function CombatCheck(t) return t.inCombat ~= nil end
local function AggroCheck(t) return t.threatSituation == 3 or t.threatSituation == 2 or t.threatSituation == 1 end
local function QuestBoss(t) return t.isQuest ~= nil end
local function ShowDirection(t) return t.direction ~= nil end
local function PlayerInRaid(t) return DB.GetAttribute("units", "player", "raid") ~= nil end
local function PowerManaCheck(t) return t.powerType == 0 or t.powerType == "MANA" end
local function PowerNotManaCheck(t) return t.powerType ~= 0 and t.powerType ~= "MANA" end
local function FullPower(t) return t.power == t.powerMax end
local function NotFullPower(t) return t.power ~= t.powerMax end
local function StanceCheck(t, unit, id) return t.stance == id end
local function ResurrectionCheck(t) return t.resurrection ~= nil end

local function DebuffCheck(t) -- check if player unit has an active debuff of the standard types
	if t.isPlayer and t.auras and t.auras.debuffs then
		local at = t.auras.debuffs.atypes
		if at and (at.Curse or at.Disease or at.Magic or at.Poison) then return true end
	end
	return false
end

local function ClassCondition1(t)
	local class = Puffin.myClass
	local sp = (class == "DRUID") and "Lifebloom" or ((class == "PRIEST") and "Power Word: Shield" or nil)
	if sp then return AuraLookup(t, sp, true, true) ~= nil end
	return false
end
	
local function ClassCondition2(t)
	local class = Puffin.myClass
	local sp = (class == "DRUID") and "Rejuvenation" or ((class == "PRIEST") and "Renew" or nil)
	if sp then return AuraLookup(t, sp, true, true) ~= nil end
	return false
end

local function ClassCondition3(t)
	local class = Puffin.myClass
	local sp = (class == "DRUID") and "Regrowth" or ((class == "PRIEST") and "Prayer of Mending" or nil)
	if sp then return AuraLookup(t, sp, true, true) ~= nil end
	return false
end

local function ClassCondition4(t)
	local class = Puffin.myClass
	local sp = (class == "DRUID") and "Wild Growth" or ((class == "PRIEST") and "Glyph of Prayer of Healing" or nil)
	if sp then return AuraLookup(t, sp, true, true) ~= nil end
	return false
end

local function ClassCondition5(t) 
	local class = Puffin.myClass
	if class == "DRUID" then return (AuraLookup(t, "Innervate", true, false) ~= nil) or (AuraLookup(t, "Thorns", true, false) ~= nil) end
	if class == "PRIEST" then return (AuraLookup(t, "Guardian Spirit", true, false) ~= nil) or (AuraLookup(t, "Fear Ward", true, false) ~= nil) end
	return false
end

local function ClassCondition6(t)
	local class = Puffin.myClass
	if class == "DRUID" and t.auras and t.auras.buffs and t.auras.buffs.cache and (t.auras.buffs.cache.Rejuvenation or t.auras.buffs.cache.Regrowth) then
		local start, duration = GetSpellCooldown("Swiftmend"); if not start or start == 0 or duration <= 1.5 then return true end
	end
	if class == "PRIEST" and t.auras and t.auras.debuffs and t.auras.debuffs.cache and t.auras.debuffs.cache["Weakened Soul"] then return true end
	return false
end

local function ClassCondition7(t)
	local class = Puffin.myClass
	if t.auras and t.auras.buffs and t.auras.buffs.cache then
		local c = t.auras.buffs.cache
		if class == "DRUID" and not (c["Mark of the Wild"] or c["Blessing of Kings"]) then return true end
		if class == "PRIEST" and not (c["Power Word: Fortitude"] or c["Commanding Shout"] or c["Blood Pact"]) then return true end
	end
	return false
end

-- Table of condition providers
local unit_condition = {
	is_resting = PlayerResting, is_pvp = PlayerPVP, raid_icon = RaidIconCheck, assigned_role = AssignedRole, in_combat = CombatCheck, aggro = AggroCheck,
	is_leader = Leader, is_master_looter = MasterLooter, ready_check_active = ReadyCheckActive, combo_points = HasComboPoints, quest_boss = QuestBoss,
	show_direction = ShowDirection, in_raid = PlayerInRaid, power_mana = PowerManaCheck, power_not_mana = PowerNotManaCheck, debuff = DebuffCheck,
	full_power = FullPower, not_full_power = NotFullPower, resurrection = ResurrectionCheck, stance = StanceCheck,
	class_condition1 = ClassCondition1, class_condition2 = ClassCondition2, class_condition3 = ClassCondition3, class_condition4 = ClassCondition4,
	class_condition5 = ClassCondition5, class_condition6 = ClassCondition6, class_condition7 = ClassCondition7, 
}

-- Icon providers
local icons = {
	Horde = { texture = [[Interface\TargetingFrame\UI-PVP-Horde]], texcoord = { 0.08, 0.58, 0.045, 0.545 }, },
	Alliance = { texture = [[Interface\TargetingFrame\UI-PVP-Alliance]], texcoord = { 0.07, 0.58, 0.06, 0.57 }, },
	FFA = { texture = [[Interface\TargetingFrame\UI-PVP-FFA]], texcoord = { 0.05, 0.605, 0.015, 0.57 }, },
	TANK = { texture = [[Interface\LFGFrame\UI-LFG-ICON-PORTRAITROLES]], texcoord = { 0, 19/64, 22/64, 41/64 }, },
	HEALER = { texture = [[Interface\LFGFrame\UI-LFG-ICON-PORTRAITROLES]], texcoord = { 20/64, 39/64, 1/64, 20/64 }, },
	DAMAGER = { texture = [[Interface\LFGFrame\UI-LFG-ICON-PORTRAITROLES]], texcoord = { 20/64, 39/64, 22/64, 41/64 }, },
	Leader = { texture = [[Interface\GroupFrame\UI-Group-LeaderIcon]], texcoord = { 0.1, 0.84, 0.14, 0.88 }, },
	Looter = { texture = [[Interface\GroupFrame\UI-Group-MasterLooter]], texcoord = { 0.15, 0.9, 0.15, 0.9 }, },
	rest = { texture = [[Interface\CharacterFrame\UI-StateIcon]], texcoord = { 0.09, 0.43, 0.08, 0.42 } },
	ready = [[Interface\RAIDFRAME\ReadyCheck-Ready]],
	notready = [[Interface\RAIDFRAME\ReadyCheck-NotReady]],
	waiting = [[Interface\RAIDFRAME\ReadyCheck-Waiting]],
	quest = [[Interface\TargetingFrame\PortraitQuestBadge]],
	Raid1 = [[Interface\TargetingFrame\UI-RaidTargetingIcon_1]],
	Raid2 = [[Interface\TargetingFrame\UI-RaidTargetingIcon_2]],
	Raid3 = [[Interface\TargetingFrame\UI-RaidTargetingIcon_3]],
	Raid4 = [[Interface\TargetingFrame\UI-RaidTargetingIcon_4]],
	Raid5 = [[Interface\TargetingFrame\UI-RaidTargetingIcon_5]],
	Raid6 = [[Interface\TargetingFrame\UI-RaidTargetingIcon_6]],
	Raid7 = [[Interface\TargetingFrame\UI-RaidTargetingIcon_7]],
	Raid8 = [[Interface\TargetingFrame\UI-RaidTargetingIcon_8]],
	arrow = "Interface\\Addons\\Puffin\\Images\\Arrow",
	indicator = "Interface\\AddOns\\Puffin\\Images\\Indicator",
}

local function PVPIcon(t) if t.pvpFFA then return icons.FFA end if t.pvp and t.faction then return icons[t.faction] end	return nil end
local function RaidIcon(t) if t.raidTarget then return icons["Raid" .. t.raidTarget] end return nil end
local function RoleIcon(t) if t.role ~= "NONE" then return icons[t.role] end return nil end
local function ReadyIcon(t) if t.ready then return icons[t.ready] end return nil end

-- Table of icon providers, returns a table with fields texture and texcoord
local unit_icon = {
	pvp = PVPIcon, raid_icon = RaidIcon, role_icon = RoleIcon, leader_icon = icons.Leader, master_looter_icon = icons.Looter,
	ready_icon = ReadyIcon, rest_icon = icons.rest, arrow_icon = icons.arrow, indicator = icons.indicator,
}

function Adapter_UnitProviders.OnEnable()
	DB = Puffin.Component("Data") -- link to data management functions
	assert(DB, EF("Puffin.Adapter_UnitProviders", "data component missing"))
	
	local bright_up, bright_down, sat_down = -0.2, -0.9, -0.25
	for k, v in pairs(cc) do local t = CopyColor(v); t.r, t.g, t.b = AdjustColor(t.r, t.g, t.b, sat_down, bright_down); ccb[k] = t end
	for k, t in pairs(cc) do t.r, t.g, t.b = AdjustColor(t.r, t.g, t.b, 0, bright_up) end
	for k, v in pairs(PowerBarColor) do local t = CopyColor(v); t.r, t.g, t.b = AdjustColor(t.r, t.g, t.b, 0, bright_up); pc[k] = t end; 
	for k, v in pairs(PowerBarColor) do local t = CopyColor(v); t.r, t.g, t.b = AdjustColor(t.r, t.g, t.b, sat_down, bright_down); pcb[k] = t end
	
	local mana_color = DB.GetAttribute("defaults", "colors", "mana")
	if mana_color then
		pc[0] = mana_color; local t = CopyColor(mana_color); t.r, t.g, t.b = AdjustColor(t.r, t.g, t.b, sat_down, bright_down); pcb[0] = t
	end
	
	for k, v in pairs(hc) do local t = CopyColor(v); t.r, t.g, t.b = AdjustColor(t.r, t.g, t.b, sat_down, bright_down); hcb[k] = t end
	for k, t in pairs(hc) do t.r, t.g, t.b = AdjustColor(t.r, t.g, t.b, 0, bright_up) end
	
	for i = 1, 50 do
		local t = AdjustHealthColor((i - 1) / 49)
		hct[i] = CopyColor(t)
		t.r, t.g, t.b = AdjustColor(t.r, t.g, t.b, sat_down, bright_down)
		hctb[i] = t
		t = hct[i]; t.r, t.g, t.b = AdjustColor(t.r, t.g, t.b, 0, bright_up)
	end

	for i = 1, 5 do combo_text[i] = "" .. combo_point_text; for j = 2, i do combo_text[i] = combo_text[i] .. " " .. combo_point_text end end
	
	DB.AllocateDatabase("providers", false)
	DB.SetRecord("providers", "unit_text", unit_text)
	DB.SetRecord("providers", "unit_color", unit_color)
	DB.SetRecord("providers", "unit_number", unit_number)
	DB.SetRecord("providers", "unit_condition", unit_condition)
	DB.SetRecord("providers", "unit_icon", unit_icon)
	DB.SetRecord("providers", "color_palette", cp)
end

Puffin.RegisterComponent("Adapter_UnitProviders", Adapter_UnitProviders, "ADAPTER", 1)
