-- Puffin: A Tojaso FowlPlay Project
-- Created August 20, 2011

Puffin = {}
local Puffin = Puffin

-- Error handling support
function Puffin.ErrorFormat(component, errmsg) return "Puffin (" .. component .. "): " .. errmsg end

-- Component types and base priorities
local componentTypes = { MODEL = 100, ADAPTER = 200, VIEW = 300, WIDGET = 400 }

-- Component management
local components = {}
local sorted_components = {}
local function sorter(a, b) return (components[a]._priority or 0) < (components[b]._priority or 0) end

local function SortComponents()
	table.wipe(sorted_components)
	local i = 0; for k in pairs(components) do sorted_components[i] = k; i = i + 1 end
	table.sort(sorted_components, sorter)
end

-- Callback to all components, in order of increasing priority
local function ComponentCallbacks(key)
	for _, k in pairs(sorted_components) do
		local c = components[k]
		if c then local cb = c[key]; if cb then cb(key) end end
	end
end

-- Register a component as represented by a table, generally done when addon is loaded
-- Optionally include a component type and priority to be used to control the order of component callbacks
function Puffin.RegisterComponent(name, ctable, ctype, priority)
	if name and ctable then
		components[name] = ctable
		if ctype then
			ctable._type = ctype
			ctable._priority = (componentTypes[ctype] or 0) + (priority or 0)
		end
		SortComponents()
	end
end

-- Unregister and disable a component
function Puffin.UnregisterComponent(name)
	if name and components[name] then
		ComponentCallbacks("OnDisable"); components[name] = nil
		SortComponents()
	end
end

-- Return the component's registered table
function Puffin.Component(name)
	if name then return components[name] end
	return nil
end

function Puffin.ADDON_LOADED(e, name) if name == "Puffin" then ComponentCallbacks("OnInitialize") end end
function Puffin.PLAYER_LOGIN() ComponentCallbacks("OnEnable") end
function Puffin.PLAYER_ENTERING_WORLD() ComponentCallbacks("OnUpdate") end -- happens before entering combat so allows initialization
function Puffin.PLAYER_LOGOUT() ComponentCallbacks("OnExit") end

-- Memory management
local tablePool = {} -- pool of available tables
function Puffin.AllocateTable() local b = next(tablePool); if b then tablePool[b] = nil else b = {} end return b end
function Puffin.ReleaseTable(b) if b then table.wipe(b); tablePool[b] = true end return nil end

-- Callback handling
local callbacks = {} -- table of tables for named callbacks
local called = false -- set to true when any callback is triggered 

-- Register a callback for a specified name (multiple callbacks can be registered for a given name)
-- Callbacks are generally functions but you can set any non-nil value
function Puffin.RegisterCallback(name, callback)
	if name and callback then
		local c = callbacks[name]
		if not c then c = Puffin.AllocateTable(); callbacks[name] = c end
		table.insert(c, callback)
	end
end

-- Unregister a callback for a specified name, releasing the name itself when the last callback is unregistered
function Puffin.UnregisterCallback(name, callback)
	if name and callback then
		local c = callbacks[name]
		if c then
			for k, cb in pairs(c) do if cb == callback then c[k] = nil; break end end -- only unregister one instance of the callback
			if not next(c) then callbacks[name] = Puffin.ReleaseTable(c) end
		end
	end
end

-- Trigger any callbacks for the specified name, if any, passing the name as argument plus any additional parameters that are passed in.
-- As a side effect, set "called" to true when any callbacks are registered for the name.
function Puffin.Callback(name, ...)
	if name and callbacks[name] then
		for k, cb in pairs(callbacks[name]) do if type(cb) == "function" then cb(name, ...) end end
		called = true
	end
end

-- Return true if any callbacks are registered for the specified name
function Puffin.CallbackRegistered(name) return name and callbacks[name] and next(callbacks[name]) end

-- Timers
local timers = {}
local timed = false --set to true when any timer is triggered
function Puffin.StartTimer(name, timeLeft) timers[name] = GetTime() + timeLeft end

-- Check for expiring timers, triggering callbacks when found and, as a side effect, setting "timed" to true when any are found
local function CheckTimers()
	local now = GetTime()
	for name, t in pairs(timers) do if t < now then timers[name] = nil; Puffin.Callback(name); timed = true end end
	return false
end

-- Update management
local updateTime, refreshTime, throttleCount = 0, 0, 0 -- time since last update and refresh, number of throttled updates
local updateThrottle, refreshThrottle, maxThrottle = 0.1, 0.05, 5 -- max 10 updates and/o 20 refreshs per second, min update every 0.5 second

local forced = false -- set to force an immediate update, obviously you don't want to do this very often
function Puffin.ForceUpdate() forced = true end

-- Handle update calls generated by the game, throttle to minimize performance impact, turn them into either OnUpdate or OnRefresh calls
-- depending on whether any events have triggered callbacks that may need to be processed by components
local function UpdateHandler(e, elapsed)
	Puffin.inLockdown = InCombatLockdown() -- global indicating that no secure functions can be called during this update
	CheckTimers() -- check for any expiring timers
	updateTime = updateTime + elapsed; refreshTime = refreshTime + elapsed
	if forced or (updateTime >= updateThrottle) then
		throttleCount = throttleCount + 1; if throttleCount == maxThrottle then throttleCount = 0; timed = true end
		if called or timed then ComponentCallbacks("OnUpdate") else ComponentCallbacks("OnRefresh") end
		updateTime = 0; refreshTime = 0; called = false; timed = false
	elseif refreshTime >= refreshThrottle then
		ComponentCallbacks("OnRefresh"); refreshTime = 0
	end
end

-- Event handling
local eventFrame = CreateFrame("Frame", nil, WorldFrame)
local function EventHandler(f, event, ...) Puffin.Callback(event, ...) end

Puffin.RegisterCallback("ADDON_LOADED", Puffin.ADDON_LOADED)
Puffin.RegisterCallback("PLAYER_LOGIN", Puffin.PLAYER_LOGIN)
Puffin.RegisterCallback("PLAYER_ENTERING_WORLD", Puffin.PLAYER_ENTERING_WORLD)
Puffin.RegisterCallback("PLAYER_LOGOUT", Puffin.PLAYER_LOGOUT)

eventFrame:SetScript("OnEvent", EventHandler)
eventFrame:SetScript("OnUpdate", UpdateHandler)
eventFrame:RegisterAllEvents()
