-- Puffin: A Tojaso FowlPlay Project
-- View component for a unit frame that is assembled from an arbitrary collection of widgets, as defined
-- by configuration records containing defaults for each class of widget plus specifics for each widget instance
-- in the view, and updated in response to changes in an assigned unit record

-- Defaults: default attribute values for each class of widget used by the view
-- Widgets: templates that specify exactly which widgets instances go into the view

-- Widget templates include the following attributes:
-- class - widget class name [c]
-- priority - if true then configure this widget before others [c]
-- parent - parent widget for scale and alpha adjustments [c]
-- anchor - widget to anchor to (default: anchor to UIParent plus view's offsetX and offsetY) [c]
-- anchorPoint - if anchor defined then point on anchor to position relative to (default: BOTTOMLEFT) [c]
-- relativePoint - point on this widget to position relative to (default: BOTTOMLEFT) [c]
-- offsetX - horizontal offset from anchor (default: 0) [c]
-- offsetY - vertical offset from anchor (default: 0) [c]
-- width - width of widget [c]
-- height - height of widget [c]
-- scale - factor to scale widget's display size [c]
-- depth - relative to view's depth (range: 0 .. 9) [c]
-- attributes - table of widget-specific attributes applied during configuration [c]
-- providers -- table of data providers used during updates to get attribute values [c]
-- alpha - opacity for the widget [d]
-- hidden - hide the widget [d]

-- View API:
-- View_UnitFrame.Create() - a new instance of a unit frame view
-- View_UnitFrame:Release() - release an instance of a unit frame view, recycling it for future use
-- View_UnitFrame:Configure() - configure the unit frame based on current defaults and template records
-- View_UnitFrame:Update(record, changed) - update the unit frames if the unit's data record has changed

-- View attributes:
-- defaults - table with default settings, including those for each class of widget [c]
-- widgets - table of widget templates actually used to construct this unit frame [c]
-- unit - table containing data for the unit associated with this unit frame [c]
-- width - width of interactive area for the unit frame [c]
-- height - height of interactive area for the unit frame [c]
-- offsetX - horizontal offset from BOTTOMLEFT of overall screen to BOTTOMLEFT of interactive area (default: 0) [c]
-- offsetY - vertical offset from BOTTOMLEFT of overall screen to BOTTOMLEFT of interactive area (default: 0) [c]
-- depth - display depth for this view [c]
-- scale - factor to scale view's display size [c]
-- hidden - hide the view [d]

local View_UnitFrame = {}
local EF = Puffin.ErrorFormat
local WU, DB, LSM
local recycledViews = {}
local viewNumber = 0
local classDepth = 1 -- mininum depth for view's of this class

-- Return an initialized instance of a unit frame view
function View_UnitFrame.Create(noUnit)
	local n, view = next(recycledViews) -- get any available recycled widget
	if n then
		recycledViews[n] = nil
	else
		view = {}
		viewNumber = viewNumber + 1
		view._widgets = {}
	end
	if not noUnit then view._unit = WU.Create() end -- optionally create a secure unit button widget
	return view
end

-- Release a view when no longer needed, recycle after clearing attributes and hiding widgets associated with it
function View_UnitFrame:Release()
	if self._unit then WU.Release(self._unit); self._unit = nil end -- release the secure unit button widget, if any
	for k, w in pairs(self._widgets) do w._class.Release(w); self._widgets[k] = nil end -- release the other widgets used by this view
	for k in pairs(self) do if (k ~= "_widgets") then self[k] = nil end end -- clear attributes for this view
	table.insert(recycledViews, self)
end

-- Check an attribute to see if it refers to special media
local function CheckMedia(v)
	if type(v) == "table" and v[1] == "#LSM" then v = LSM:Fetch(v[2], v[3]) end -- check for LibSharedMedia references
	return v
end

-- Configure a widget based on its descriptor
local function ConfigureWidget(uf, k, w, offsetDepth)
	local t = uf._widgets[k] -- check if already allocated this widget
	if not t then -- need to create an instance of the widget
		local class = Puffin.Component(w.class)
		assert(class, EF("View_UnitFrame", "unrecognized class"))
		t = class.Create(); t._class = class; t._className = w.class; uf._widgets[k] = t
	end
	local x, y, a = (w.offsetX or 0) + ((w.relativeX or 0) * uf.width), (w.offsetY or 0) + ((w.relativeY or 0) * uf.height), w.anchor
	if a then a = (uf._widgets[a] or nil) else x = x + (uf.offsetX or 0); y = y + (uf.offsetY or 0) end
	t._class.SetPosition(t, w.anchorPoint or uf.anchorPoint or "BOTTOMLEFT", a, w.relativePoint or uf.relativePoint or "BOTTOMLEFT", x, y, w.anchorRegion)
	if w.parent then t._class.SetParent(t, uf._widgets[w.parent]) end
	t.width = (w.width or 0) + ((w.relativeWidth or 0) * uf.width); t.height = (w.height or 0) + ((w.relativeHeight or 0) * uf.height)
	t.depth = (w.depth or 0) + offsetDepth; if w.alpha then t.alpha = w.alpha end; if w.scale then t.scale = w.scale end
	if w.attributes then for i, v in pairs(w.attributes) do t[i] = CheckMedia(v) end end -- widget specific attributes
	local defaults = DB.GetRecord("widget_defaults", uf.defaults)
	if defaults and defaults[w.class] then -- default attributes for this widget class
		for i, v in pairs(defaults[w.class]) do if t[i] == nil then t[i] = CheckMedia(v) end end
	end
	t._class.Configure(t)
end

-- Configure all widgets associated with this view
function View_UnitFrame:Configure()
	assert(self.width and self.width > 0, EF("View_UnitFrame", "undefined width"))
	assert(self.height and self.height > 0, EF("View_UnitFrame", "undefined height"))
	assert(self.defaults, EF("View_UnitFrame", "undefined defaults"))
	assert(self.templates, EF("View_UnitFrame", "undefined templates"))
	assert(self.unit, EF("View_UnitFrame", "undefined unit"))
	local offsetDepth = ((self.depth or 0) + classDepth) * 10
	if self._unit then
		local u = self._unit
		WU.SetPosition(u, self.anchorPoint or "BOTTOMLEFT", nil, self.relativePoint or "BOTTOMLEFT", self.offsetX or 0, self.offsetY or 0) -- position unit widget
		u.unit = self.unit -- name of unit for this unit frame
		u.depth = offsetDepth + 9 -- use highest depth reserved for this unit frame view
		u.width = self.width; u.height = self.height; u.clique = self.clique
		WU.Configure(u)
	end

	local templates = DB.GetRecord("widget_templates", self.templates)
	local widgets = DB.GetRecord("widgets", self.widgets)
	if widgets and templates then
		for i = 1, 10 do -- priority order for configuring (required for parent and position dependencies)
			for _, w in pairs(widgets) do -- each widget has unique name as key into templates table
				local template = templates[w]
				if template and (template.priority == i) then ConfigureWidget(self, w, template, offsetDepth) end
			end
		end
	end
end

-- Update all widgets associated with this view, using widget-specific providers to access the data record
function View_UnitFrame:Update(changed, record, unit)
	local templates = DB.GetRecord("widget_templates", self.templates)
	local widgets = DB.GetRecord("widgets", self.widgets)
	if widgets and templates then
		for _, w in pairs(widgets) do -- each widget has unique name as key into templates table
			local template = templates[w]
			local t = self._widgets[w]
			assert(t, EF("View_UnitFrame", "update called for invalid widget " .. w))
			t.hidden = self.hidden
			if not self.hidden and changed and template.providers then -- call providers for widget specific data attributes
				for i, p in pairs(template.providers) do
					if record and type(p) == "table" then
						local v = DB.GetAttribute("providers", p.class, p.name)
						if type(v) == "function" and not p.no_function then v = v(record, unit, p.param) end
						if type(v) == "table" and not p.no_table and v[i] then for j, k in pairs(v) do t[j] = k end else t[i] = v end
					elseif i == "condition" then  -- special case to test for existence of the unit
						t[i] = (p == "exists") and (record ~= nil) or false
					end
				end
			end
			t._class.Update(t)
		end
	end
end

function View_UnitFrame.OnEnable()
	DB = Puffin.Component("Data") -- link to data management functions
	assert(DB, EF("View_UnitFrames", "Data component not found"))
	WU = Puffin.Component("Widget_Unit") -- handle for unit widget class
	assert(WU, EF("View_UnitFrames", "Widget_Unit not found"))
	LSM = LibStub("LibSharedMedia-3.0")
	assert(LSM, EF("Adapter_Configuration", "LibSharedMedia not found"))
end

Puffin.RegisterComponent("View_UnitFrame", View_UnitFrame, "VIEW", 0)
