-- Puffin: A Tojaso FowlPlay Project
-- Widget to conditionally show an icon

-- Custom attributes:
-- texture - path to rectangle's texture (default: color swatch) [d]
-- texcoord - table containing icon's texcoord settings (default: { 0, 1, 0, 1 }) [d]
-- color - color to apply to icon (default: none) [d]
-- condition - only show the icon when true (default: false) [d]
-- rotation - amount of rotation to apply to the texture in radians (default: none) [d]
-- startAnimation - animation to play when icon is shown: fade, pulse, delay, splash (default: none) [d]
-- endAnimation - animation to play when icon is hidden: fade, pulse, delay, splash (default: none) [d]

local Widget_Icon = {}
local EF = Puffin.ErrorFormat
local recycledIcons = {}
local activeIcons = {}
local iconNumber = 0 -- used to create unique frame names
local animationPool = {} -- pool of available animations
local animations = {} -- active animations

-- Initialize and return an animation for a widget's icon texture
-- Supported styles: fade, delay, pulse, splash
local function IconAnimation(widget, style)
	local t = widget._display
	local tex = t.texture:GetTexture(); if not tex then return end
	local ULx, ULy, LLx, LLy, URx, URy, LRx, LRy = t.texture:GetTexCoord()
	local offsetX, offsetY = t.container:GetLeft(), t.container:GetBottom()
	local b = next(animationPool)
	if b then animationPool[b] = nil else
		b = {} -- initializae a new animation
		b.frame = CreateFrame("Frame", nil, UIParent)
		b.frame:SetFrameLevel(t.container:GetFrameLevel() + 9)
		b.texture = b.frame:CreateTexture(nil, "ARTWORK") -- texture for the texture to be animated	
		b.anim = b.frame:CreateAnimationGroup()
		b.anim:SetLooping("NONE")
		b.grow = b.anim:CreateAnimation("Scale"); b.grow:SetOrigin('CENTER', 0, 0); b.grow:SetOrder(1)
		b.shrink = b.anim:CreateAnimation("Scale"); b.shrink:SetOrigin('CENTER', 0, 0); b.shrink:SetOrder(2)
		b.alpha = b.anim:CreateAnimation("Alpha")
	end
	local w, h = t.texture:GetSize()
	b.frame:ClearAllPoints(); b.frame:SetPoint("BOTTOMLEFT", nil, "BOTTOMLEFT", offsetX, offsetY); b.frame:SetSize(w, h); b.frame:Show()
	b.texture:SetTexture(tex); b.texture:SetTexCoord(ULx, ULy, LLx, LLy, URx, URy, LRx, LRy); b.texture:SetAlpha(widget.alpha or 1)
	local c = widget.color; if c then b.texture:SetVertexColor(c.r, c.g, c.b) else b.texture:SetVertexColor(1, 1, 1) end
	b.texture:ClearAllPoints(); b.texture:SetAllPoints(b.frame); b.texture:Show()
	local animationTime = 1
	if style == "fade" then -- 0.65 second fade
		b.grow:SetScale(1, 1); b.grow:SetDuration(0.1) -- disable grow
		b.shrink:SetScale(1, 1); b.shrink:SetDuration(0.1) -- disable shrink
		b.alpha:SetChange(-1); b.alpha:SetDuration(0.65); b.alpha:SetSmoothing("IN"); b.alpha:SetEndDelay(5); b.alpha:SetOrder(1)
	elseif style == "pulse" then -- 0.5 second fade with scale grow/shrink
		b.grow:SetScale(3, 3); b.grow:SetDuration(0.25)
		b.shrink:SetScale(-3, -3); b.shrink:SetDuration(0.25)
		b.alpha:SetChange(-1); b.alpha:SetDuration(0.25); b.alpha:SetSmoothing("IN"); b.alpha:SetEndDelay(5); b.alpha:SetOrder(2)
	elseif style == "delay" then -- 2 second fade
		b.grow:SetScale(1, 1); b.grow:SetDuration(0.1) -- disable grow
		b.shrink:SetScale(1, 1); b.shrink:SetDuration(0.1) -- disable shrink
		b.alpha:SetChange(-1); b.alpha:SetDuration(2); b.alpha:SetSmoothing("IN"); b.alpha:SetEndDelay(5); b.alpha:SetOrder(1)
		animationTime = 3
	else -- must be "splash" 0.65 sec fade with scale flyout
		b.grow:SetScale(3, 3); b.grow:SetDuration(0.65)
		b.shrink:SetScale(1, 1); b.shrink:SetDuration(0.1) -- disable shrink
		b.alpha:SetChange(-1); b.alpha:SetDuration(0.65); b.alpha:SetSmoothing("IN"); b.alpha:SetEndDelay(5); b.alpha:SetOrder(1)
	end
	b.anim:Stop(); b.anim:Play()
	b.endTime = GetTime() + animationTime -- stop after one second
	table.insert(animations, b)
end

-- Update active animations, recycling when they are complete
function Widget_Icon:OnUpdate()
	local now = GetTime()
	for k, b in pairs(animations) do
		if now > b.endTime then
			b.anim:Pause(); animations[k] = nil; animationPool[b] = true
			b.frame:ClearAllPoints(); b.texture:ClearAllPoints(); b.frame:Hide(); b.texture:Hide()
		end
	end
end

-- Create a new icon widget object
function Widget_Icon.Create()
	local n, icon = next(recycledIcons) -- get any available recycled widget
	if n then
		recycledIcons[n] = nil
	else
		icon = {}
		iconNumber = iconNumber + 1
		local t = {}; icon._display = t -- table of display objects preserved when widget is recycled
		local frameName = "Puffin_Icon" .. tostring(iconNumber) -- generate unique name for widget's frame
		t.container = CreateFrame("Frame", frameName .. "Container", UIParent)
		t.container:SetFrameStrata("MEDIUM")
		t.texture = t.container:CreateTexture(nil, "ARTWORK")
	end
	table.insert(activeIcons, icon) -- keep track of widgets so can animate if necessary
	return icon
end

-- Release widget when no longer needed, recycle object after clearing attributes and hiding display objects associated with it
function Widget_Icon:Release()
	for k in pairs(self) do if k ~= "_display" then self[k] = nil end end
	self._display.container:Hide()
	for k, v in pairs(activeIcons) do if v == self then activeIcons[k] = nil end end
	table.insert(recycledIcons, self)
end

-- Set display position for the widget relative to another widget or the overall screen
function Widget_Icon:SetPosition(anchor, anchorWidget, relative, offsetX, offsetY, region)
	local t, w = self._display, UIParent
	if anchorWidget then w = region and anchorWidget._display[region] or anchorWidget._display.container end
	t.container:ClearAllPoints(); t.container:SetPoint(anchor, w, relative, offsetX, offsetY)
end

-- Set the parent widget (used for managing display heirarchy)
function Widget_Icon:SetParent(widget)
	local t, w = self._display, widget._display
	if w and w.container then w = w.container else w = UIParent end
	t.container:SetParent(w)
	if not self.depth then t.container:SetFrameLevel(w:GetFrameLevel() + 1) end
end

-- Configure all the widget's display objects based on current configuration attribute values (or defaults)
function Widget_Icon:Configure()
	assert(self.width and self.width > 0, EF("Widget_Icon", "undefined width"))
	assert(self.height and self.height > 0, EF("Widget_Icon", "undefined height"))
	local w, h = self.width, self.height
	if w > h then w = h elseif h > w then h = w end -- make sure texture is square
	local t = self._display
	t.texture:SetAlpha(1); t.texture:SetSize(w, h)
	t.container:SetScale(self.scale or 1); t.container:SetSize(self.width, self.height); 
	t.texture:Hide(); t.container:Hide()
end

-- Update the widget's display objects based on current data attribute values (or defaults)
function Widget_Icon:Update()
	local t = self._display
	if self.hidden or (self.condition == false) then
		if t.container:IsShown() and self.endAnimation then IconAnimation(self, self.endAnimation) end
		t.texture:ClearAllPoints(); t.texture:SetPoint("BOTTOMLEFT", t.container, "BOTTOMRIGHT"); t.texture:SetPoint("TOPRIGHT", t.container, "TOPLEFT")
		t.texture:Hide(); t.container:Hide() -- setting size to 0 supports automatically positioning rows of icons
	else
		if not t.container:IsShown() and self.startAnimation then IconAnimation(self, self.startAnimation) end
		if self.texture then t.texture:SetTexture(self.texture) else t.texture:SetTexture(1, 1, 1) end
		if self.texcoord then t.texture:SetTexCoord(unpack(self.texcoord)) else t.texture:SetTexCoord(0, 1, 0, 1) end
		t.texture:ClearAllPoints(); t.texture:SetPoint("BOTTOMLEFT", t.container, "BOTTOMLEFT")
		if self.rotation then t.texture:SetRotation(self.rotation); self._rotation = self.rotation elseif self._rotation then self._rotation = nil end
		
		local d = self.depth; if d then t.container:SetFrameLevel(d) end
		local c = self.color; if c then t.texture:SetVertexColor(c.r, c.g, c.b) else t.texture:SetVertexColor(1, 1, 1) end
		t.container:SetAlpha(self.alpha or 1); t.texture:Show(); t.container:Show()
	end
end

-- Animate rotations of the texture
function Widget_Icon:Animate()
	if not (self.hidden or (self.condition == false)) then
		local dir = Puffin.mouseoverDirection
		if dir and self._rotation and dir ~= self._rotation then
			local t = self._display
			if t.texture then t.texture:SetRotation(dir); self._rotation = dir end
		end
	end
end

function Widget_Icon:OnRefresh() for _, w in pairs(activeIcons) do Widget_Icon.Animate(w) end end

-- Register this widget as a component so it can be found by views
Puffin.RegisterComponent("Widget_Icon", Widget_Icon, "WIDGET", 0)
